package com.mortals.xhx.opencv.utils;



import lombok.extern.slf4j.Slf4j;
import org.bytedeco.javacv.Java2DFrameUtils;
import org.bytedeco.opencv.opencv_core.*;
import org.bytedeco.opencv.opencv_core.Point;


import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.image.BufferedImage;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Base64;

import static org.bytedeco.opencv.global.opencv_core.CV_8UC3;
import static org.bytedeco.opencv.global.opencv_core.flip;
import static org.bytedeco.opencv.global.opencv_imgproc.*;
import static org.bytedeco.opencv.global.opencv_imgcodecs.*;

/**
 * @author ZYW
 * @date 2023-12-21 10:09
 */
@Slf4j
public class OpencvHelper {

    /**
     * 图像缩放
     * @param width 限制宽度
     * @param height 限制高度
     * @return 缩放后的mat
     */
    public static BufferedImage ImageZoom(Mat sourceImage, int width, int height){
        //创建缩放后的图像
        Size newSize = new Size(width, height);
        Mat outImage = new Mat();
        Mat.zeros(sourceImage.size(), sourceImage.type());
        //执行缩放操作
        resize(sourceImage, outImage, newSize, 0, 0, INTER_LINEAR);
        //保存图片
//        Imgcodecs.imwrite("path/to/save/resized_image.jpg", resizedImage);
        return matToBufferedImage(outImage);
    }

    /**
     * 图像缩放
     * @param width 限制宽度
     * @param height 限制高度
     * @return 缩放后的mat
     */
    public static Mat ImageZoomMat(Mat sourceImage, int width, int height){
        //创建缩放后的图像
        Size newSize = new Size(width, height);
        Mat outImage = new Mat();
        Mat.zeros(sourceImage.size(), sourceImage.type());
        //执行缩放操作
        resize(sourceImage, outImage, newSize, 0, 0, INTER_LINEAR);
        //保存图片
//        Imgcodecs.imwrite("path/to/save/resized_image.jpg", resizedImage);
        return outImage;
    }


    /**
     * 图像旋转
     * @param sourceImage 输入图片mat
     * @param rote 旋转角度90、180、270
     * @return 旋转后的mat
     */
    public static BufferedImage ImageRotate(Mat sourceImage, int rote){
        // 获取图像的尺寸
        int width = sourceImage.cols();
        int height = sourceImage.rows();
        Mat outImage = new Mat();

        // 创建一个旋转矩阵，指定旋转角度（以度为单位）
        Mat rotationMatrix = getRotationMatrix2D(new Point2f(width / 2, height / 2), rote, 1);

        // 使用旋转矩阵旋转图像，并将结果存储在rotatedImage中
        warpAffine(sourceImage, outImage, rotationMatrix, new Size(width, height));
        return matToBufferedImage(outImage);
    }

    /**
     * 图像旋转
     * @param sourceImage 输入图片mat
     * @param rote 旋转角度90、180、270
     * @return 旋转后的mat
     */
    public static Mat ImageRotateMat(Mat sourceImage, int rote){
        if(rote == 0){
            return sourceImage;
        }else {
            // 获取图像的尺寸
            int width = sourceImage.cols();
            int height = sourceImage.rows();
            Mat outImage = new Mat();

            // 创建一个旋转矩阵，指定旋转角度（以度为单位）
            Mat rotationMatrix = getRotationMatrix2D(new Point2f(width / 2, height / 2), rote, 1);

            // 使用旋转矩阵旋转图像，并将结果存储在rotatedImage中
            warpAffine(sourceImage, outImage, rotationMatrix, new Size(width, height));
            return outImage;
        }
    }

    /**
     * 图片翻转
     * @param inputPath 输入图片
     * @param flip  1表示水平翻转，0表示垂直翻转，-1表示同时进行水平和垂直翻转
     * @return
     */
    public static BufferedImage ImageFlip(String inputPath,int flip)
    {
        //读取图像
        Mat sourceImage = imread(inputPath);
        if (sourceImage.empty()) {
            System.err.println("Could not read the image");
            return null;
        }
        Mat outImage = new Mat();
        // 使用Imgproc.flip进行上下翻转，1表示水平翻转，0表示垂直翻转，-1表示同时进行水平和垂直翻转
        flip(sourceImage, outImage, flip);
        return matToBufferedImage(outImage);
    }

    /**
     * 图片自动裁切
     * @param sourceImage 输入图片
     * @return
     */
    public static BufferedImage ImageCutAuto(Mat sourceImage)
    {
//        // 转为灰度图像
//        Mat gray = new Mat();
//        cvtColor(sourceImage,  gray, COLOR_BGR2GRAY);
//
//        // 二值化处理
//        Mat binary = new Mat();
//        threshold(gray, binary, 0, 255, THRESH_BINARY |THRESH_OTSU);
//
//        // 边缘检测
//        Mat edges = new Mat();
//        Canny(binary, edges, 50, 150);
//
//        // 寻找轮廓
//        MatVector matVector = new MatVector();
//        Mat hierarchy = new Mat();
//        findContours(binary, matVector, hierarchy, RETR_TREE, CHAIN_APPROX_SIMPLE);
//
//        // 找出最大的轮廓（假设这是图片内容）
//        double maxArea = 0;
//        int maxContourIndex = 0;
//        for (int i = 0; i < matVector.size(); i++) {
//            double area = contourArea(matVector.get(i));
//            if (area > maxArea) {
//                maxArea = area;
//                maxContourIndex = i;
//            }
//        }
//
//        // 根据最大轮廓裁剪图像
//        Rect rect = boundingRect(matVector.get(maxContourIndex));
//        Mat outImage = new Mat(sourceImage, rect);
//        return matToBufferedImage(outImage);
        BufferedImage bufferedImage = matToBufferedImage(sourceImage);

        // 转为灰度图像
        Mat gray = new Mat();
        cvtColor(sourceImage,  gray, COLOR_BGR2GRAY);

        // 二值化处理
        Mat binary = new Mat();
        threshold(gray, binary, 80, 255, THRESH_BINARY |THRESH_OTSU);

        //降噪 - 中值滤波
        Mat blur = new Mat();
        medianBlur(binary,blur,3);

        //降噪 - 高斯滤波
        Mat gauss = new Mat();
        GaussianBlur(blur,gauss,new Size(3,3), 0);

        //  膨胀和腐蚀生成轮廓
        Mat element = getStructuringElement(MORPH_ELLIPSE, new Size(3, 3));
        Mat dilate = new Mat();
        dilate(gauss, dilate, element);//膨胀：将前景物体变大，理解成将图像断开裂缝变小
        Mat erode = new Mat();
        erode(dilate, erode, element);//腐蚀：将前景物体变小，理解成将图像断开裂缝变大

        // 边缘检测
        Mat edges = new Mat();
        Canny(erode, edges, 30, 90);

        // 寻找轮廓
        MatVector matVector = new MatVector();
        Mat hierarchy = new Mat();
        findContours(edges, matVector, hierarchy, RETR_TREE, CHAIN_APPROX_SIMPLE);

        // 找出最大的轮廓（假设这是图片内容）
        double maxArea = 0;
        int maxContourIndex = 0;
        for (int i = 0; i < matVector.size(); i++) {
            double area = contourArea(matVector.get(i));
            if (area > maxArea) {
                maxArea = area;
                maxContourIndex = i;
            }
        }

        //计算最小外接旋转矩形
        RotatedRect rectRotate = minAreaRect(matVector.get(maxContourIndex));

        // 计算裁剪后的边界点
        Point[] points = new Point[4];
        points(rectRotate,points);

        //获取中心点和大小
        Point2f center = rectRotate.center();
        Size2f size2f = rectRotate.size();
        Size size = new Size((int)size2f.width(), (int)size2f.height());


//        //计算透视变换矩阵
//        Mat M = getPerspectiveTransform();
//        Mat outImage = new Mat();
//        warpPerspective(sourceImage, outImage, M, size);

        // 根据最大轮廓裁剪图像
        Rect rect = boundingRect(matVector.get(maxContourIndex));
        Mat outImage = new Mat(sourceImage,rect);


        return matToBufferedImage(outImage);
    }

    /**
     * 获取矩形旋转后的坐标
     * @param pt
     */
    public static void points(RotatedRect rect, Point[] pt) {
        double _angle = rect.angle() * Math.PI / 180.0;
        double b = Math.cos(_angle) * 0.5;
        double a = Math.sin(_angle) * 0.5;
        pt[0] = new Point((int) (rect.center().x() - a * rect.size().height() - b * rect.size().width()), (int) (rect.center().y() + b * rect.size().height() - a * rect.size().width()));
        pt[1] = new Point((int) (rect.center().x() + a * rect.size().height() - b * rect.size().width()), (int) (rect.center().y() - b * rect.size().height() - a * rect.size().width()));
        pt[2] = new Point((int) (2.0 * rect.center().x() - pt[0].x()), (int) (2.0 * rect.center().y() - pt[0].y()));
        pt[3] = new Point((int) (2.0 * rect.center().x() - pt[1].x()), (int) (2.0 * rect.center().y() - pt[1].y()));
    }

    /**
     * 图片自动裁切
     * @param sourceImage 输入图片
     * @return
     */
    public static Rectangle cutAutoGetAngele(Mat sourceImage)
    {
        // 转为灰度图像
        Mat gray = new Mat();
        cvtColor(sourceImage,  gray, COLOR_BGR2GRAY);

        // 二值化处理
        Mat binary = new Mat();
        threshold(gray, binary, 0, 255, THRESH_BINARY |THRESH_OTSU);

        //降噪 - 中值滤波
        Mat blur = new Mat();
        medianBlur(binary,blur,3);

        //降噪 - 高斯滤波
        Mat gauss = new Mat();
        GaussianBlur(blur,gauss,new Size(3,3), 0);

        //  膨胀和腐蚀生成轮廓
        Mat element = getStructuringElement(MORPH_ELLIPSE, new Size(3, 3));
        Mat dilate = new Mat();
        dilate(gauss, dilate, element);//膨胀：将前景物体变大，理解成将图像断开裂缝变小
        Mat erode = new Mat();
        erode(dilate, erode, element);//腐蚀：将前景物体变小，理解成将图像断开裂缝变大

        // 边缘检测  加了边缘监测 框就会一直闪烁
        Mat edges = new Mat();
        Canny(erode, edges, 30, 90);

        // 寻找轮廓
        MatVector matVector = new MatVector();
        Mat hierarchy = new Mat();
        findContours(erode, matVector, hierarchy, RETR_TREE, CHAIN_APPROX_SIMPLE);


        // 找出最大的轮廓（假设这是图片内容）
        double maxArea = 0;
        int maxContourIndex = 0;
        for (int i = 0; i < matVector.size(); i++) {
            double area = contourArea(matVector.get(i));
            if (area > maxArea) {
                maxArea = area;
                maxContourIndex = i;
            }
        }

        //计算最小外接旋转矩形
        RotatedRect rectRotate = minAreaRect(matVector.get(maxContourIndex));

        // 计算裁剪后的边界点
        Point[] points = new Point[4];
        points(rectRotate,points);

//        Scalar color = new Scalar(0, 0, 255,5);
//        line(sourceImage,points[0],points[1],color);
//        line(sourceImage,points[1],points[2],color);
//        line(sourceImage,points[2],points[3],color);
//        line(sourceImage,points[3],points[0],color);


        // 根据最大轮廓裁剪图像
        Rect rect = boundingRect(matVector.get(maxContourIndex));
        Rectangle rectangle = new Rectangle(rect.x(),rect.y(),rect.width(),rect.height());
        return rectangle;
    }

    /**
     * 图片按照划定矩形裁切
     * @param sourceImage 输入图片
     * @return
     */
    public static BufferedImage ImageCutByRect(Mat sourceImage,Rect rect)
    {
        // 根据划定轮廓裁剪图像
        Mat outImage = new Mat(sourceImage, rect);
        return matToBufferedImage(outImage);
    } 


        // 将Mat转换为BufferedImage的方法
    public static BufferedImage matToBufferedImage(Mat matrix) {
        return Java2DFrameUtils.toBufferedImage(matrix);

    }

    // 将BufferedImage转换为Mat的方法
    public static Mat imageToMat(BufferedImage bi) {
        return Java2DFrameUtils.toMat(bi);
    }

    // 将BufferedImage转换为base64的方法
    public static String imageToBase64(BufferedImage bufferedImage) {
        // 创建ByteArrayOutputStream
        ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();

        // 将BufferedImage写入到ByteArrayOutputStream中
        try {
            ImageIO.write(bufferedImage, "jpg", byteArrayOutputStream);
        } catch (IOException e) {
            return "";
        }

        // 将数据编码为Base64字符串
        String base64String = Base64.getEncoder().encodeToString(byteArrayOutputStream.toByteArray());
        return "data:image/jpg;base64,"+base64String;
    }

    // 根据三个点计算中间那个点的夹角   pt1 pt0 pt2
    private static double getAngle(Point pt1, Point pt2, Point pt0)
    {
        double dx1 = pt1.x() - pt0.x();
        double dy1 = pt1.y() - pt0.y();
        double dx2 = pt2.x() - pt0.x();
        double dy2 = pt2.y() - pt0.y();
        return (dx1*dx2 + dy1*dy2)/Math.sqrt((dx1*dx1 + dy1*dy1)*(dx2*dx2 + dy2*dy2) + 1e-10);
    }

    // 找到最大的正方形轮廓
    private static int findLargestSquare(MatVector matVector) {
        double maxArea = 0;
        int maxContourIndex = 0;
        for (int i = 0; i < matVector.size(); i++) {
            double area = contourArea(matVector.get(i));
            if (area > maxArea) {
                maxArea = area;
                maxContourIndex = i;
            }
        }

        return maxContourIndex;
    }

    // 点到点的距离
    private static double getSpacePointToPoint(Point p1, Point p2) {
        double a = p1.x() - p2.x();
        double b = p1.y() - p2.y();
        return Math.sqrt(a * a + b * b);
    }

    // 两直线的交点
    private static Point computeIntersect(double[] a, double[] b) {
        if (a.length != 4 || b.length != 4)
        {throw new ClassFormatError();}
        double x1 = a[0], y1 = a[1], x2 = a[2], y2 = a[3], x3 = b[0], y3 = b[1], x4 = b[2], y4 = b[3];
        double d = ((x1 - x2) * (y3 - y4)) - ((y1 - y2) * (x3 - x4));
        if (d != 0) {
            double  x = ((x1 * y2 - y1 * x2) * (x3 - x4) - (x1 - x2) * (x3 * y4 - y3 * x4)) / d;
            double  y = ((x1 * y2 - y1 * x2) * (y3 - y4) - (y1 - y2) * (x3 * y4 - y3 * x4)) / d;
            Point pt = new Point((int) x,(int) y);

            return pt;
        }
        else{
            return new Point(-1, -1);
        }
    }

    // 对多个点按顺时针排序
    private static void sortCorners(ArrayList<Point> corners) {
        if (corners.size() == 0) return;
        Point p1 = corners.get(0);
        int index = 0;
        for (int i = 1; i < corners.size(); i++) {
            Point point = corners.get(i);
            if (p1.x() > point.x()) {
                p1 = point;
                index = i;
            }
        }

        corners.set(index, corners.get(0));
        corners.set(0, p1);

        Point lp = corners.get(0);
        for (int i = 1; i < corners.size(); i++) {
            for (int j = i + 1; j < corners.size(); j++) {
                Point point1 = corners.get(i);
                Point point2 = corners.get(j);
                if ((point1.y()-lp.y()*1.0)/(point1.x()-lp.x())>(point2.y()-lp.y()*1.0)/(point2.x()-lp.x())) {
                    Point temp = new Point(point1.x(),point1.y());
                    corners.set(i, corners.get(j));
                    corners.set(j, temp);
                }
            }
        }
    }

}
